# Tests of in-memory notebooks
#
# Copyright (C) 2016 Simon Dobson
# 
# Licensed under the GNU General Public Licence v.2.0
#

from epyc import *

import six
import unittest
import os


class SampleExperiment(Experiment):
    '''A very simple experiment that adds up its parameters.'''

    def do( self, param ):
        total = 0
        for k in param:
            total = total + param[k]
        return dict(total = total)

    
class LabNotebookTests(unittest.TestCase):

    def testEmptyNotebook( self ):
        '''Test creating an empty notebook'''
        nb = LabNotebook("test")
        self.assertEqual(nb.name(), "test")
        self.assertFalse(nb.isPersistent())
        
    def testAddingResult( self ):
        '''Test adding and retrieving a result'''
        nb = LabNotebook()

        e = SampleExperiment()
        params = dict(a  = 1, b = 2)
        rc = e.set(params).run()

        nb.addResult(rc)

        self.assertNotEqual(nb.resultsFor(params), [])
        res = nb.results()
        self.assertEqual(len(res), 1)
        self.assertEqual(res[0][Experiment.RESULTS]['total'], params['a'] + params['b'])

    def testPermuteParameters( self ):
        '''Test parameters are normalised properly.'''
        nb = LabNotebook()

        e = SampleExperiment()
        params1 = dict(a  = 1, b = 2)
        rc = e.set(params1).run()

        nb.addResult(rc)
        self.assertNotEqual(nb.resultsFor(params1), [])

        params2 = dict(b = 2, a  = 1)
        self.assertNotEqual(nb.resultsFor(params2), [])

        self.assertEqual(nb.resultsFor(params1), nb.resultsFor(params2))

    def testLatestResults( self ):
        '''Check that latest result works properly.'''
        nb = LabNotebook()

        e = SampleExperiment()
        params1 = dict(a  = 1, b = 2)
        rc = e.set(params1).run()

        six.assertCountEqual(self, nb.resultsFor(params1), [])
        self.assertEqual(nb.latestResultsFor(params1), None)

        nb.addResult(rc)
        self.assertEqual(len(nb.resultsFor(params1)), 1)
        self.assertEqual(nb.latestResultsFor(params1), rc)
        
    def testAddingPendingResult( self ):
        '''Test adding, finalising, and retrieving a pending result'''
        nb = LabNotebook()

        e = SampleExperiment()
        params = dict(a  = 1, b = 2)
        rc = e.set(params).run()
        
        nb.addPendingResult(params, 1)
        six.assertCountEqual(self, nb.resultsFor(params), [])
        self.assertEqual(len(nb.results()), 0)
        six.assertCountEqual(self, nb.pendingResults(), [ 1 ])

        nb.addResult(rc, 1)
        self.assertEqual(len(nb.resultsFor(params)), 1)
        self.assertEqual((nb.resultsFor(params))[0], rc)
        self.assertEqual(len(nb.results()), 1)
        self.assertEqual(len(nb.pendingResults()), 0)

    def testCancellingPendingResult( self ):
        '''Test cancelling a pending result'''
        nb = LabNotebook()

        params = dict(a  = 1, b = 2)
        nb.addPendingResult(params, 1)
        six.assertCountEqual(self, nb.resultsFor(params), [])
        self.assertEqual(len(nb.results()), 0)
        six.assertCountEqual(self, nb.pendingResults(), [ 1 ])

        nb.cancelPendingResult(1)
        six.assertCountEqual(self, nb.resultsFor(params), [])
        self.assertEqual(len(nb.results()), 0)
        self.assertEqual(len(nb.pendingResults()), 0)

    def testCancellingAllPendingResults( self ):
        '''Test cancelling of all pending result'''
        nb = LabNotebook()

        params1 = dict(a  = 1, b = 2)
        params2 = dict(a  = 1, b = 3)
        nb.addPendingResult(params1, 1)
        nb.addPendingResult(params2, 2)
        self.assertEqual(nb.resultsFor(params1), [])
        self.assertEqual(nb.resultsFor(params2), [])
        self.assertEqual(len(nb.results()), 0)
        self.assertIn(1, nb.pendingResults())
        self.assertIn(2, nb.pendingResults())

        nb.cancelAllPendingResults()
        six.assertCountEqual(self, nb.resultsFor(params1), [])
        self.assertEqual(len(nb.results()), 0)
        self.assertEqual(len(nb.pendingResults()), 0)
    
    def testRealAndPendingResults( self ):
        '''Test a sequence of real and pending results'''
        nb = LabNotebook()

        e = SampleExperiment()
        
        params1 = dict(a  = 1, b = 2)
        rc1 = e.set(params1).run()
        params2 = dict(a  = 10, b = 12)
        rc2 = e.set(params2).run()
        params3 = dict(a  = 45, b = 11)
        rc3 = e.set(params3).run()

        nb.addResult(rc1)
        self.assertEqual((nb.resultsFor(params1))[0], rc1)
        six.assertCountEqual(self, nb.resultsFor(params2), [])
        self.assertEqual(len(nb.results()), 1)
        six.assertCountEqual(self, nb.pendingResults(), [])
        
        nb.addPendingResult(params2, 2)
        self.assertEqual((nb.resultsFor(params1))[0], rc1)
        six.assertCountEqual(self, nb.resultsFor(params2), [])
        self.assertEqual(len(nb.results()), 1)
        six.assertCountEqual(self, nb.pendingResults(), [ 2 ])
        
        nb.addPendingResult(params3, 3)
        self.assertEqual((nb.resultsFor(params1))[0], rc1)
        six.assertCountEqual(self, nb.resultsFor(params2), [])
        six.assertCountEqual(self, nb.resultsFor(params3), [])
        self.assertEqual(len(nb.results()), 1)
        six.assertCountEqual(self, nb.pendingResults(), [ 2, 3 ])

        nb.addResult(rc2, 2)
        self.assertEqual((nb.resultsFor(params1))[0], rc1)
        self.assertEqual((nb.resultsFor(params2))[0], rc2)
        self.assertEqual(nb.resultsFor(params3), [])
        self.assertEqual(len(nb.results()), 2)
        six.assertCountEqual(self, nb.pendingResults(), [ 3 ])

        nb.cancelPendingResult(3)
        self.assertEqual((nb.resultsFor(params1))[0], rc1)
        self.assertEqual((nb.resultsFor(params2))[0], rc2)
        six.assertCountEqual(self, nb.resultsFor(params3), [])
        self.assertEqual(len(nb.results()), 2)
        six.assertCountEqual(self, nb.pendingResults(), [])

        nb.addPendingResult(params3, 3)
        self.assertEqual((nb.resultsFor(params1))[0], rc1)
        self.assertEqual((nb.resultsFor(params2))[0], rc2)
        six.assertCountEqual(self, nb.resultsFor(params3), [])
        self.assertEqual(len(nb.results()), 2)
        six.assertCountEqual(self, nb.pendingResults(), [ 3 ])

        nb.cancelPendingResult(3)
        self.assertEqual((nb.resultsFor(params1))[0], rc1)
        self.assertEqual((nb.resultsFor(params2))[0], rc2)
        six.assertCountEqual(self, nb.resultsFor(params3), [])
        self.assertEqual(len(nb.results()), 2)
        six.assertCountEqual(self, nb.pendingResults(), [])

    def testCancellingAllPendingResults( self ):
        '''Test all results get cancelled properly.'''
        nb = LabNotebook()

        e = SampleExperiment()
        
        params1 = dict(a  = 1, b = 2)
        rc1 = e.set(params1).run()
        params2 = dict(a  = 10, b = 12)
        rc2 = e.set(params2).run()
        rc3 = e.set(params2).run()

        nb.addPendingResult(params1, 1)
        nb.addPendingResult(params2, 2)
        nb.addPendingResult(params2, 3)
        six.assertCountEqual(self, nb.pendingResultsFor(params1), [ 1 ])
        six.assertCountEqual(self, nb.pendingResultsFor(params2), [ 2, 3 ])
        six.assertCountEqual(self, nb.pendingResults(), [ 1, 2, 3 ])

        nb.cancelPendingResultsFor(params2)
        six.assertCountEqual(self, nb.pendingResultsFor(params1), [ 1 ])
        six.assertCountEqual(self, nb.pendingResultsFor(params2), [])
        six.assertCountEqual(self, nb.pendingResults(), [ 1 ])

        nb.cancelAllPendingResults()
        six.assertCountEqual(self, nb.pendingResultsFor(params1), [])
        six.assertCountEqual(self, nb.pendingResultsFor(params2), [])
        six.assertCountEqual(self, nb.pendingResults(), [])
        six.assertCountEqual(self, nb.results(), [])

    def testAddResultList( self ):
        '''Test adding several results at once.'''
        nb = LabNotebook()

        e = SampleExperiment()
        
        params1 = dict(a  = 1, b = 2)
        rc1 = e.set(params1).run()
        params2 = dict(a  = 10, b = 12)
        rc2 = e.set(params2).run()
        rc3 = e.set(params2).run()

        nb.addResult([ rc1, rc2, rc3 ])
        self.assertEqual(nb.latestResultsFor(params1), rc1)
        six.assertCountEqual(self, nb.resultsFor(params2), [ rc2, rc3 ])
        self.assertEqual(nb.latestResultsFor(params2), rc3)
        
        
    def testDataFrame( self ):
        '''Test creating a pandas DataFrame'''
        nb = LabNotebook()

        e = SampleExperiment()
        params1 = dict(a  = 1, b = 2)
        rc1 = e.set(params1).run()
        params2 = dict(a  = 10, b = 12)
        rc2 = e.set(params2).run()
        params3 = dict(a  = 45, b = 11)
        #rc4 = e.set(params2).run()

        nb.addResult(rc1)
        nb.addResult(rc2)
        nb.addPendingResult(params3, 1)
        #nb.addResult(rc4)

        df = nb.dataframe()

        self.assertTrue(len(df['a']), 2)
        self.assertTrue(len(df['b']), 2)
        self.assertTrue(len(df['total']), 2)
        self.assertEqual(df[df['a'] == 1]['b'].iloc[0], 2)
        self.assertEqual(df[df['b'] == 12]['a'].iloc[0], 10)
        self.assertEqual(df[(df['a'] == 10) & (df['b'] == 12)]['total'].iloc[0], 10 + 12)
        
    def testNoJob( self ):
        '''Test we can't remove a non-existent job.'''
        nb = LabNotebook()

        params = dict(a  = 1, b = 2)
        nb.addPendingResult(params, 1)
        six.assertCountEqual(self, nb.resultsFor(params), [])
        self.assertEqual(len(nb.results()), 0)
        six.assertCountEqual(self, nb.pendingResults(), [ 1 ])

        with self.assertRaises(KeyError):
            nb.cancelPendingResult(2)
        
    def testNoJobTwice( self ):
        '''Test we can't remove the same job twice.'''
        nb = LabNotebook()

        params = dict(a  = 1, b = 2)
        nb.addPendingResult(params, 1)
        six.assertCountEqual(self, nb.resultsFor(params), [])
        self.assertEqual(len(nb.results()), 0)
        six.assertCountEqual(self, nb.pendingResults(), [ 1 ])
        
        nb.cancelPendingResult(1)
        self.assertEqual(len(nb.results()), 0)
        six.assertCountEqual(self, nb.pendingResults(), [])

        with self.assertRaises(KeyError):
            nb.cancelPendingResult(1)
        



# Simulation "lab notebook" for collecting results, persistent version using SQLite
#
# Copyright (C) 2016 Simon Dobson
# 
# Licensed under the GNU General Public Licence v.2.0
#

from epyc import *

import sqlite3
import os


class SqliteLabNotebook(LabNotebook):
    """A "laboratory notebook" recording the results of a set of
    experiments conducted across a parameter space. The intention is
    to record all the metadata necessary to re-conduct the experiment."""

    # names of significant database elements
    EXPERIMENT_METADATA_TABLE = "experiment_metadata"
    EXPERIMENT_ID = "id"
    RESULTS_ID = "job"
    
    def __init__( self, name, dbfile = ':memory:', create = False ):
        """Create an empty notebook in the given database file, or in
        memory by default.

        dbfile: name of the database. By default creates an in-memory database
        create: if True, overwrite any existing database in the named file.
                By default opens the database"""
        super(SqlliteLabNotebook, self).__init__(name)
        
        self._dbfile = dbfile
        self._connection = None

        # if file exists, open or create the database
        if os.path.isfile(self._dbfile):
            # file exists, open or create?
            if create:
                # creating, erase the file and create a new database
                os.remove(self._dbfile)
                self.open()
                self._createDatabase()
            else:
                # re-using, open the database
                self.open()
        else:
            # file doesn't exist or is in memory, create database
            self.open()
            self._createDatabase()

    def isPersistent( self ):
        """Tests whether the notebook is persistent (i.e., stored on disc).

        returns: True if notebook is persistent, False otherwise"""
        return not (self._dbfile == ':memory:')
        
    def open( self ):
        """Open the database connection."""
        if self._connection is None:
            self._connection = sqlite3.connect(self._dbfile)

    def close( self ):
        """Close the database connection."""
        if self._connection is not None:
            self._connection.close()
            self._connection = None

    def commit( self ):
        """Private method to commit changes to database."""
        self._connection.commit()
        
    def _createDatabase( self ):
        """Private method to create the SQLite database file."""

        # create experiment metadata table
        command = """
                  CREATE TABLE {tn} (
                  {k}             INT PRIMARY KEY NOT NULL,
                  START_TIME      INT             NOT NULL,
                  END_TIME        INT             NOT NULL,
                  ELAPSED_TIME    INT             NOT NULL,
                  SETUP_TIME      INT             NOT NULL,
                  EXPERIMENT_TIME INT             NOT NULL,
                  TEARDOWN_TIME   INT             NOT NULL,
                  STATUS          BOOLEAN         NOT NULL)
                  """
        self._connection.execute(command.format(tn = self.EXPERIMENT_METADATA_TABLE,
                                                k = self.EXPERIMENT_ID))
        
        # commit the changes
        self.commit()
        
    def addExperiment( self, rc ):
        """Add the results of an experiment to the notebook. The dict
        of results should be structured according to the Experiment
        class, with metadata and results stored separately. The
        metadata is placed into the experiment_metadata table,
        the results into the experiment_results table; and the two
        linked through the experiment_instances table.

        rc: the dict of experimental results"""
        pass



# Base class for experiments
#
# Copyright (C) 2016--2020 Simon Dobson
# 
# This file is part of epyc, experiment management in Python.
#
# epyc is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# epyc is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with epyc. If not, see <http://www.gnu.org/licenses/gpl.html>.

from __future__ import print_function
from datetime import datetime, timedelta
import sys
import traceback

class Experiment(object):
    """Base class for an :term:`experiment` conducted in a :term:`lab`.

    An :class:`Experiment` defines a computational experiment that can be run
    independently or (more usually) controlled from an instamnce of the
    :class:`Lab` class. Experiments should be long-lasting, able to conduct
    repeated runs at several different parameter points.

    From an experimenter's (or a lab's) perspective, an experiment
    has public methods :meth:`set` and :meth:`run`. The former sets the parameters
    for the experiment; the latter runs the experiment. producing a set
    of results that include direct experimental results and metadata on
    how the experiment ran. A single run may produce a list of result
    dicts if desired, each filled-in with the correct metadata.

    Experimental results, parameters, and metadata can be access directly
    from the :class:`Experiment` object. The class also exposes
    an indexing interface to access experimental results by name.
    """

    # Top-level structure for results
    METADATA = 'metadata'                 #: Results dict key for metadata values, mainly timing
    PARAMETERS = 'parameters'             #: Results dict key for describing the point in the parameter space the experiment ran on
    RESULTS = 'results'                   #: Results dict key for the experimental results generated at the experiment's parameter point

    # Common metadata elements reported
    START_TIME = 'start_time'             #: Metadata element for the datetime experiment started
    END_TIME = 'end_time'                 #: Metadata element for the datetime experiment ended
    ELAPSED_TIME = 'elapsed_time'         #: Metadata element for the time experiment took overall in seconds
    SETUP_TIME = 'setup_time'             #: Metadata element for the time spent on setup in seconds
    EXPERIMENT_TIME = 'experiment_time'   #: Metadata element for the time spent on experiment itself in seconds
    TEARDOWN_TIME = 'teardown_time'       #: Metadata element for the time spent on teardown in seconds
    STATUS = 'status'                     #: Metadata element that will be True if experiment completed successfully, False otherwise
    EXCEPTION = 'exception'               #: Metadata element containing the exception thrown if experiment failed
    TRACEBACK = 'traceback'               #: Metadata element containing the traceback from the exception (as a string)

    
    def __init__( self ):
        """Create a new experiment."""
        self._metadata = dict()
        self._parameters = None
        self._results = None


    # ---------- Configuration ----------

    def set(self, params):
        """Set the parameters for the experiment, returning the
        now-configured experiment. Be sure to call this base method when overriding.

        :param params: the parameters
        :returns: the experiment"""
        if self._parameters is not None:
            self.deconfigure()
        self.configure(params)
        return self

    def configure( self, params ):
        """Configure the experiment for the given parameters.
        The default stores the parameters for later use. Be sure
        to call this base method when overriding.

        :param params: the parameters"""
        self._parameters = params

    def deconfigure( self ):
        """De-configure the experiment prior to setting new parameters.
        Default removes the parameters. Default does nothing."""
        self._parameters = None


    # ---------- Running the experiment ----------

    def setUp( self, params ):
        """Set up the experiment. Default does nothing.

        :param params: the parameters of the experiment"""
        pass

    def tearDown( self ):
        """Tear down the experiment. Default does nothing."""
        pass

    def do( self, params ):
        """Do the body of the experiment. This should be overridden
        by sub-classes.

        params: a dict of parameters for the experiment
        returns: a dict of experimental results"""
        raise NotImplementedError('do()')

    def report( self, params, meta, res ):
        """Return a properly-structured dict of results. The default returns a dict with
        results keyed by :attr:`Experiment.RESULTS`, the data point in the parameter space
        keyed by :attr:`Experiment.PARAMETERS`, and timing and other metadata keyed
        by :attr:`Experiment.METADATA`. Overriding this method can be used to record extra
        metadata values, but be sure to call the base method as well.

        :param params: the parameters we ran under
        :param meta: the metadata for this run
        :param res: the direct experimental results from do()
        :returns: a :term:`results dict`"""
        rc = dict()
        rc[self.PARAMETERS] = params.copy()
        rc[self.METADATA] = meta.copy()
        rc[self.RESULTS] = res
        return rc

    def run( self ):
        """Run the experiment, using the parameters set using :meth:`set`.
        A "run" consists of calling :meth:`setUp`, :meth:`do`, and :meth:`tearDown`,
        followed by collecting and storing (and returning) the
        experiment's results. If running the experiment raises an
        exception, that will be returned in the metadata along with
        its traceback to help with experiment debugging.

        :returns: a :term:`results dict`"""
        
        # perform the experiment protocol
        params = self.parameters()
        self._metadata = dict()
        self._results = None
        res = None
        doneSetupTime = doneExperimentTime = doneTeardownTime = None
        try:
            # do the phases in order, recording the wallclock times at each phase
            startTime = datetime.now()
            self.setUp(params)
            doneSetupTime = datetime.now()
            res = self.do(params)
            doneExperimentTime = datetime.now() 
            self.tearDown()
            doneTeardownTime = datetime.now() 
            
            # record the various timings
            self._metadata[self.START_TIME] = startTime
            self._metadata[self.END_TIME] = doneTeardownTime
            self._metadata[self.ELAPSED_TIME] = (doneTeardownTime - startTime).total_seconds()
            self._metadata[self.SETUP_TIME] = (doneSetupTime - startTime).total_seconds()
            self._metadata[self.EXPERIMENT_TIME] = (doneExperimentTime - doneSetupTime).total_seconds()
            self._metadata[self.TEARDOWN_TIME] = (doneTeardownTime - doneExperimentTime).total_seconds()
            
            # set the success flag
            self._metadata[self.STATUS] = True
        except Exception as e:
            print("Caught exception in experiment: {e}".format(e = e))

            # grab the traceback before we do anything else
            #_, _, tb = sys.exc_info()
            tb = traceback.format_exc()
            
            # decide on the cleanup actions that need doing
            if (doneSetupTime is not None) and (doneExperimentTime is None):
                # we did the setup and then failed in the experiment, so
                # we need to do the teardown
                try:
                    self.tearDown()
                except:
                    pass
                
            # set the failure flag and record the exception
            # (there will be no timing information recorded)
            self._metadata[self.STATUS] = False
            self._metadata[self.EXCEPTION] = e
            self._metadata[self.TRACEBACK] = tb
                
        # report the results
        self._results = res
        return self.report(params,
                           self._metadata,
                           res)


    # ---------- Accessing results ----------

    def __getitem__( self, k ):
        """Return the given element of the experimental results. This only
        gives access to the experimental results, *not* to the parameters
        or metadata.

        :param k: the result key
        :returns: the value
        :raises: KeyError if there is no such result"""
        if self._results is None:
            raise KeyError(k)
        else:
            return (self.experimentalResults())[k]
    
    def success( self ):
        """Test whether the experiment has been run successfully. This will
        be False if the experiment hasn't been run, or if it's been run and
        failed.

        :returns: ``True`` if the experiment has been run successfully"""
        if self.STATUS in self.metadata().keys():
            return (self.metadata())[self.STATUS]
        else:
            return False

    def failed(self):
        '''Test whether an experiment failed. This will be True if the experiment has
        been run and has failed, which means that there will be an exception and traceback
        information stored in the metadata.

        :returns: ``True`` if the experiment has failed'''
        if self.STATUS not in self.metadata().keys():
            return False
        else:
            return not (self.metadata())[self.STATUS]

    def results( self ):
        """Return a complete results dict. Only really makes sense for
        recently-executed experimental runs.

        :returns: the results dict"""
        return self.report(self.parameters(),
                           self.metadata(),
                           self.experimentalResults())

    def experimentalResults( self ):
        """Return the experimental results from our last run. This will
        be None if we haven't been run, or if we ran and failed.

        :returns: the experimental results or ``None``"""
        return self._results
    
    def parameters( self ):
        """Return the current experimental parameters, which will
        be None if none have been given by a call to set()

        :returns: the parameters,"""
        return self._parameters

    def metadata( self ):
        """Return the metadata we collected at out last execution, which
        will be None if we've not been executed and an empty dict if
        we're mid-run (i.e., if this method is called from do() for
        whatever reason).

        :returns: the metadata"""
        return self._metadata
    
    
        
    



# Simulation "lab notebook" for collecting results, JSON version
#
# Copyright (C) 2016 Simon Dobson
# 
# This file is part of epyc, experiment management in Python.
#
# epyc is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# epyc is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with epyc. If not, see <http://www.gnu.org/licenses/gpl.html>.

import epyc

import os
import json
import numpy
from datetime import datetime
import dateutil.parser
import types


class MetadataEncoder(json.JSONEncoder):
    """Add support for encoding Python datetime objects within
    JSON objects, using the standard ISO string representation.
    (Plus a few other minor tweaks to get things to work more smoothly.)"""

    def default( self, o ):
        """If o is a datetime object, convert it to an ISO string. If it is an
        exception, convert it to a string. If it is a numpy int, coerce it to
        a Python int.

        :param o: the field to serialise
        :returns: a string encoding of the field"""
        if isinstance(o, datetime):
            # date/time, return an ISO formatted string
            return o.isoformat()
        else:
            if isinstance(o, Exception):
                # exception, stringify it
                return str(o)
            else:
                if isinstance(o, numpy.integer):
                    # numpy ints inherit differently between Python 2 and 3
                    # see https://bugs.python.org/issue24313
                    return int(o)
                else:
                    # everything else, pass through
                    return super(MetadataEncoder, self).default(o)


class JSONLabNotebook(epyc.LabNotebook):
    """A lab notebook that persists intself to a JSON file. This is
    the most basic kind of persistent notebook, readable by
    virtually any tooling.

    Using JSON presents some disadvantages, as not all types can be
    represented. Specifically, exceptions from the metadata of failed
    experiments (with the :attr:`Experiment.EXCEPTION`)
    will be saved as strings (for the exception).
    We also need to convert `datetime` objects to ISO-format strings
    when saving.
    """

    def __init__( self, name, create = False, description = None ):
        """Create a new JSON notebook, using the notebook's name
        as the JSON file. If this file exists, it will be opened
        and loaded unless create is True, in which case it will
        be erased.

        :param name: JSON file to persist the notebook to
        :param create: if True, erase existing file (defaults to False)
        :param description: free text description of the notebook"""
        super(epyc.JSONLabNotebook, self).__init__(name, description)

        # check for the file already existing
        if os.path.isfile(self.name()):
            # file exists, do we load it or create into it?
            if create:
                # empty the file
                with open(self.name(), 'w') as f:
                    f.write('')

                # preserve any description we were passed
                self._description = description
            else:
                # load notebook from file
                self._load(self.name())

    def isPersistent( self ):
        """Return True to indicate the notebook is persisted to a JSON file.

        :returns: True"""
        return True

    def commit( self ):
        """Persist to disc."""
        self._save(self.name())
        
    def _load( self, fn ):
        """Retrieve the notebook from the given file.

        :param fn: the file name"""

        # if file is empty, create an empty notebook
        if os.path.getsize(fn) == 0:
            self._description = None
            self._results = dict()
            self._pending = dict()
        else:
            # load the JSON object
            with open(fn, "r") as f:
                s = f.read()

                # parse back into appropriate variables
                j = json.loads(s)
                self._description = j['description']
                self._pending = dict(j['pending'])
                self._results = j['results']

                # perform any post-load patching
                self.patch()

    def _patchDatetimeMetadata( self, res, mk ):
        """Private method to patch an ISO datetime string to a datetime object
        for metadata key mk.

        :param res: results dict
        :param mk: metadata key"""
        t = res[epyc.Experiment.METADATA][mk]
        res[epyc.Experiment.METADATA][mk] = dateutil.parser.parse(t)
        
    def patch( self ):
        """Patch the results dict. The default processes the :attr:`Experiment.START_TIME`
        and :attr:`Experiment.END_TIME` metadata fields back into Python `datetime` objects
        from ISO strings. This isn't strictly necessary, but it makes notebook
        data structure more Pythonic."""

        for k in self._results.keys():
            ars = self._results[k]
            for res in ars:
                if isinstance(res, dict) and res[epyc.Experiment.METADATA][epyc.Experiment.STATUS]:
                    # a successful, non-pending result, patch its timing metadata
                    self._patchDatetimeMetadata(res, epyc.Experiment.START_TIME)
                    self._patchDatetimeMetadata(res, epyc.Experiment.END_TIME)

    def _save( self, fn ):
        """Persist the notebook to the given file.

        :param fn: the file name"""

        # create JSON object
        j = json.dumps({ 'description': self.description(),
                         'pending': self._pending,
                         'results': self._results },
                       indent = 4,
                       cls = MetadataEncoder)

        # write to file
        with open(fn, 'w') as f:
            f.write(j)
            

    



# Tests of cluster-driven lab class
#
# Copyright (C) 2016 Simon Dobson
# 
# Licensed under the GNU General Public Licence v.2.0
#

from epyc import *

import unittest
import numpy
import time
import os
import subprocess
from tempfile import NamedTemporaryFile

# set limit low for testing purposes
ClusterLab.WaitingTime = 10


class SampleExperiment(Experiment):
    '''A very simple experiment that adds up its parameters.'''

    def do( self, param ):
        total = 0
        for k in param:
            total = total + param[k]
        return dict(total = total)

class SampleExperiment2(Experiment):
    '''Add up after waiting.'''

    def do( self, param ):
        time.sleep(2)
        total = 0
        for k in param:
            total = total + param[k]
        return dict(total = total)

    
# use the existence of an ipcontroller-client.json file in the IPython
# default profile's directory as a proxy for there being a cluster running
# that we can use for our tests
default_profile_dir = str(subprocess.check_output('ipython locate profile default', shell = True)).strip('\n')
connection_file = os.path.join(default_profile_dir, 'security/ipcontroller-client.json')
@unittest.skipUnless(os.path.isfile(connection_file),
                     "No default cluster running (no {fn})".format(fn = connection_file))
class ClusterLabTests(unittest.TestCase):

    def setUp( self ):
        '''Create a lab in which to perform tests.'''
        self._lab = ClusterLab()
        #self._lab._use_dill()
        with self._lab.sync_imports():
            import time

    def tearDown( self ):
        '''Close the conection to the cluster.'''
        self._lab.close()
        self._lab = None
        
    def testMixup( self ):
        '''Test that parameter spaces are suitably mixed, defined as not
        more than 0.5% of elements landing in their original place.'''
        n = 1000
        
        l = numpy.arange(0, n)
        self._lab._mixup(l)
        sp = [ v for v in (l == numpy.arange(0, n)) if v ]
        self.assertTrue(len(sp) <= (n * 0.005))

    def testEmpty( self ):
        '''Test that things work for an empty lab'''
        self.assertEqual(self._lab.readyFraction(), 0)
        self.assertEqual(self._lab.results(), [])
        
    def testRunExprimentSync( self ):
        '''Test running an experiment and grabbing all the results by sleeping for a while.'''
        n = 20

        r = numpy.arange(0, n)
        self._lab['a'] = r
        self._lab.runExperiment(SampleExperiment())
        time.sleep(n * 2.5 / self._lab.numberOfEngines())
        self.assertTrue(self._lab.ready())
        res = self._lab.results()
        
        # check that the whole parameter space has a result
        self.assertEqual(len(res), n)
        for p in res:
            self.assertIn(p[Experiment.PARAMETERS]['a'], r)

        # check that each result corresponds to its parameter
        for p in res:
            self.assertEqual(p[Experiment.PARAMETERS]['a'], p[Experiment.RESULTS]['total'])
            
    def testWait( self ):
        '''Test waiting for all jobs to complete.'''
        n = 20

        r = numpy.arange(0, n)
        self._lab['a'] = r
        self._lab.runExperiment(SampleExperiment())
        self.assertTrue(self._lab.wait())
        self.assertTrue(self._lab.ready())
            
    def testWaitShortTimeout( self ):
        '''Test short-timeout (and short-latency) waiting.'''
        n = 20
        self._lab.WaitingTime = 5
        
        r = numpy.arange(0, n)
        self._lab['a'] = r
        self._lab.runExperiment(SampleExperiment2())
        self.assertFalse(self._lab.wait(timeout = 5))
        self.assertFalse(self._lab.ready())
        self.assertTrue(self._lab.wait())
        self.assertTrue(self._lab.ready())

    def testRunExprimentAsync( self ):
        '''Test running an experiment and check the results come in piecemeal.'''
        n = 20

        r = numpy.arange(0, n)
        self._lab['a'] = r
        self._lab.runExperiment(SampleExperiment())

        # watch results coming in
        f = 0.0
        while f < 1:
            f1 = self._lab.readyFraction()
            #print self._lab._availableResults(), f1
            self.assertTrue(f1 >= f)
            f = f1
        self.assertTrue(self._lab.ready())
        self.assertEqual(self._lab.notebook().numberOfPendingResults(), 0)
        res = self._lab.results()
        
        # check that the whole parameter space has a result
        self.assertEqual(len(res), n)
        for p in res:
            self.assertIn(p[Experiment.PARAMETERS]['a'], r)

        # check that each result corresponds to its parameter
        for p in res:
            self.assertEqual(p[Experiment.PARAMETERS]['a'], p[Experiment.RESULTS]['total'])

    def testReturnWithNoJobs( self ):
        '''Test wait() returns True when there are no jobs pending.'''
        n = 20
          
        r = numpy.arange(0, n)
        self._lab['a'] = r
        self._lab.runExperiment(SampleExperiment())
        self.assertTrue(self._lab.wait())
        
        # calling wait() again should also be true (with no delay, which we don't check for)
        self.assertTrue(self._lab.wait())

    def testCancelSomeJobs( self ):
        '''Test we can cancel some jobs while keeping the rest.'''
        n = 20
          
        r = numpy.arange(0, n)
        self._lab['a'] = r
        self._lab.runExperiment(SampleExperiment2())

        params = dict(a = int(n / 2))
        self._lab.cancelPendingResultsFor(params)
        self._lab.wait()
        self.assertEqual(self._lab.numberOfResults(), n - 1)
        self.assertEqual(self._lab.numberOfPendingResults(), 0)
        self.assertEqual(self._lab.notebook().latestResultsFor(params), None)

    def testCancelAllJobs( self ):
        '''Test we can cancel all jobs.'''
        n = 20
          
        r = numpy.arange(0, n)
        self._lab['a'] = r
        self._lab.runExperiment(SampleExperiment2())

        self._lab.cancelAllPendingResults()
        self._lab.wait()
        self.assertEqual(self._lab.numberOfResults(), 0)
        self.assertEqual(self._lab.numberOfPendingResults(), 0)

    def testAddExperiments( self ):
        '''Test we can add experiments while some are running, without locking up.'''
        n = 20

        # run the first experiment
        r = numpy.arange(0, n)
        self._lab['a'] = r
        self._lab.runExperiment(SampleExperiment2())

        # while this is waiting, run another
        r = numpy.arange(n, 2 * n)
        self._lab['a'] = r
        self._lab.runExperiment(SampleExperiment2())

        self._lab.wait()
        self.assertEqual(self._lab.numberOfResults(), 2 * n)
        self.assertEqual(self._lab.numberOfPendingResults(), 0)

        
       
